@@ -5,13 +5,13 @@
 
     The fastest markdown parser in pure Python with renderer feature.
 
-    :copyright: (c) 2014 - 2015 by Hsiaoming Yang.
+    :copyright: (c) 2014 - 2018 by Hsiaoming Yang.
 """
 
 import re
 import inspect
 
-__version__ = '0.5.1'
+__version__ = '0.8.4'
 __author__ = 'Hsiaoming Yang <me@lepture.com>'
 __all__ = [
     'BlockGrammar', 'BlockLexer',
@@ -21,6 +21,25 @@ __all__ = [
 ]
 
 
+_key_pattern = re.compile(r'\s+')
+_nonalpha_pattern = re.compile(r'\W')
+_escape_pattern = re.compile(r'&(?!#?\w+;)')
+_newline_pattern = re.compile(r'\r\n|\r')
+_block_quote_leading_pattern = re.compile(r'^ *> ?', flags=re.M)
+_block_code_leading_pattern = re.compile(r'^ {4}', re.M)
+_inline_tags = [
+    'a', 'em', 'strong', 'small', 's', 'cite', 'q', 'dfn', 'abbr', 'data',
+    'time', 'code', 'var', 'samp', 'kbd', 'sub', 'sup', 'i', 'b', 'u', 'mark',
+    'ruby', 'rt', 'rp', 'bdi', 'bdo', 'span', 'br', 'wbr', 'ins', 'del',
+    'img', 'font',
+]
+_pre_tags = ['pre', 'script', 'style']
+_valid_end = r'(?!:/|[^\w\s@]*@)\b'
+_valid_attr = r'''\s*[a-zA-Z\-](?:\s*\=\s*(?:"[^"]*"|'[^']*'|[^\s'">]+))?'''
+_block_tag = r'(?!(?:%s)\b)\w+%s' % ('|'.join(_inline_tags), _valid_end)
+_scheme_blacklist = ('javascript:', 'vbscript:')
+
+
 def _pure_pattern(regex):
     pattern = regex.pattern
     if pattern.startswith('^'):
@@ -28,14 +47,9 @@ def _pure_pattern(regex):
     return pattern
 
 
-_key_pattern = re.compile(r'\s+')
-
-
 def _keyify(key):
-    return _key_pattern.sub(' ', key.lower())
-
-
-_escape_pattern = re.compile(r'&(?!#?\w+;)')
+    key = escape(key.lower(), quote=True)
+    return _key_pattern.sub(' ', key)
 
 
 def escape(text, quote=False, smart_amp=True):
@@ -59,23 +73,24 @@ def escape(text, quote=False, smart_amp=True):
     return text
 
 
+def escape_link(url):
+    """Remove dangerous URL schemes like javascript: and escape afterwards."""
+    lower_url = url.lower().strip('\x00\x1a \n\r\t')
+
+    for scheme in _scheme_blacklist:
+        if re.sub(r'[^A-Za-z0-9\/:]+', '', lower_url).startswith(scheme):
+            return ''
+    return escape(url, quote=True, smart_amp=False)
+
+
 def preprocessing(text, tab=4):
-    text = re.sub(r'\r\n|\r', '\n', text)
-    text = text.replace('\t', ' ' * tab)
-    text = text.replace('\u00a0', ' ')
+    text = _newline_pattern.sub('\n', text)
+    text = text.expandtabs(tab)
     text = text.replace('\u2424', '\n')
     pattern = re.compile(r'^ +$', re.M)
     return pattern.sub('', text)
 
 
-_tag = (
-    r'(?!(?:'
-    r'a|em|strong|small|s|cite|q|dfn|abbr|data|time|code|'
-    r'var|samp|kbd|sub|sup|i|b|u|mark|ruby|rt|rp|bdi|bdo|'
-    r'span|br|wbr|ins|del|img)\b)\w+(?!:/|[^\w\s@]*@)\b'
-)
-
-
 class BlockGrammar(object):
     """Grammars for block level tokens."""
 
@@ -94,7 +109,7 @@ class BlockGrammar(object):
     newline = re.compile(r'^\n+')
     block_code = re.compile(r'^( {4}[^\n]+\n*)+')
     fences = re.compile(
-        r'^ *(`{3,}|~{3,}) *(\S+)? *\n'  # ```lang
+        r'^ *(`{3,}|~{3,}) *([^`\s]+)? *\n'  # ```lang
         r'([\s\S]+?)\s*'
         r'\1 *(?:\n+|$)'  # ```
     )
@@ -103,11 +118,12 @@ class BlockGrammar(object):
     lheading = re.compile(r'^([^\n]+)\n *(=|-)+ *(?:\n+|$)')
     block_quote = re.compile(r'^( *>[^\n]+(\n[^\n]+)*\n*)+')
     list_block = re.compile(
-        r'^( *)([*+-]|\d+\.) [\s\S]+?'
+        r'^( *)(?=[*+-]|\d+\.)(([*+-])?(?:\d+\.)?) [\s\S]+?'
         r'(?:'
         r'\n+(?=\1?(?:[-*_] *){3,}(?:\n+|$))'  # hrule
         r'|\n+(?=%s)'  # def links
-        r'|\n+(?=%s)'  # def footnotes
+        r'|\n+(?=%s)'  # def footnotes\
+        r'|\n+(?=\1(?(3)\d+\.|[*+-]) )'   # heterogeneous bullet
         r'|\n{2,}'
         r'(?! )'
         r'(?!\1(?:[*+-]|\d+\.) )\n*'
@@ -135,14 +151,14 @@ class BlockGrammar(object):
             _pure_pattern(block_quote),
             _pure_pattern(def_links),
             _pure_pattern(def_footnotes),
-            '<' + _tag,
+            '<' + _block_tag,
         )
     )
     block_html = re.compile(
         r'^ *(?:%s|%s|%s) *(?:\n{2,}|\s*$)' % (
             r'<!--[\s\S]*?-->',
-            r'<(%s)[\s\S]+?<\/\1>' % _tag,
-            r'''<%s(?:"[^"]*"|'[^']*'|[^'">])*?>''' % _tag,
+            r'<(%s)((?:%s)*?)>([\s\S]*?)<\/\1>' % (_block_tag, _valid_attr),
+            r'<%s(?:%s)*?\s*\/?>' % (_block_tag, _valid_attr),
         )
     )
     table = re.compile(
@@ -185,6 +201,9 @@ class BlockLexer(object):
             rules = self.grammar_class()
 
         self.rules = rules
+        self._max_recursive_depth = kwargs.get('max_recursive_depth', 6)
+        self._list_depth = 0
+        self._blockquote_depth = 0
 
     def __call__(self, text, rules=None):
         return self.parse(text, rules)
@@ -203,14 +222,14 @@ class BlockLexer(object):
                     continue
                 getattr(self, 'parse_%s' % key)(m)
                 return m
-            return False
+            return False  # pragma: no cover
 
         while text:
             m = manipulate(text)
             if m is not False:
                 text = text[len(m.group(0)):]
                 continue
-            if text:
+            if text:  # pragma: no cover
                 raise RuntimeError('Infinite loop at: %s' % text)
         return self.tokens
 
@@ -220,9 +239,8 @@ class BlockLexer(object):
             self.tokens.append({'type': 'newline'})
 
     def parse_block_code(self, m):
-        code = m.group(0)
-        pattern = re.compile(r'^ {4}', re.M)
-        code = pattern.sub('', code)
+        # clean leading whitespace
+        code = _block_code_leading_pattern.sub('', m.group(0))
         self.tokens.append({
             'type': 'code',
             'lang': None,
@@ -260,9 +278,16 @@ class BlockLexer(object):
             'type': 'list_start',
             'ordered': '.' in bull,
         })
-        cap = m.group(0)
-        self._process_list_item(cap, bull)
+        self._list_depth += 1
+        if self._list_depth > self._max_recursive_depth:
+            self.tokens.append({'type': 'list_item_start'})
+            self.parse_text(m)
+            self.tokens.append({'type': 'list_item_end'})
+        else:
+            cap = m.group(0)
+            self._process_list_item(cap, bull)
         self.tokens.append({'type': 'list_end'})
+        self._list_depth -= 1
 
     def _process_list_item(self, cap, bull):
         cap = self.rules.list_item.findall(cap)
@@ -283,7 +308,7 @@ class BlockLexer(object):
                 pattern = re.compile(r'^ {1,%d}' % space, flags=re.M)
                 item = pattern.sub('', item)
 
-            # determin whether item is loose or not
+            # determine whether item is loose or not
             loose = _next
             if not loose and re.search(r'\n\n(?!\s*$)', item):
                 loose = True
@@ -306,11 +331,15 @@ class BlockLexer(object):
 
     def parse_block_quote(self, m):
         self.tokens.append({'type': 'block_quote_start'})
-        cap = m.group(0)
-        pattern = re.compile(r'^ *> ?', flags=re.M)
-        cap = pattern.sub('', cap)
-        self.parse(cap)
+        self._blockquote_depth += 1
+        if self._blockquote_depth > self._max_recursive_depth:
+            self.parse_text(m)
+        else:
+            # clean leading >
+            cap = _block_quote_leading_pattern.sub('', m.group(0))
+            self.parse(cap)
         self.tokens.append({'type': 'block_quote_end'})
+        self._blockquote_depth -= 1
 
     def parse_def_links(self, m):
         key = _keyify(m.group(1))
@@ -360,9 +389,9 @@ class BlockLexer(object):
         cells = cells.split('\n')
         for i, v in enumerate(cells):
             v = re.sub(r'^ *\| *| *\| *$', '', v)
-            cells[i] = re.split(r' *\| *', v)
+            cells[i] = re.split(r' *(?<!\\)\| *', v)
 
-        item['cells'] = cells
+        item['cells'] = self._process_cells(cells)
         self.tokens.append(item)
 
     def parse_nptable(self, m):
@@ -371,9 +400,9 @@ class BlockLexer(object):
         cells = re.sub(r'\n$', '', m.group(3))
         cells = cells.split('\n')
         for i, v in enumerate(cells):
-            cells[i] = re.split(r' *\| *', v)
+            cells[i] = re.split(r' *(?<!\\)\| *', v)
 
-        item['cells'] = cells
+        item['cells'] = self._process_cells(cells)
         self.tokens.append(item)
 
     def _process_table(self, m):
@@ -399,14 +428,31 @@ class BlockLexer(object):
         }
         return item
 
+    def _process_cells(self, cells):
+        for i, line in enumerate(cells):
+            for c, cell in enumerate(line):
+                # de-escape any pipe inside the cell here
+                cells[i][c] = re.sub('\\\\\|', '|', cell)
+
+        return cells
+
     def parse_block_html(self, m):
-        pre = m.group(1) in ['pre', 'script', 'style']
-        text = m.group(0)
-        self.tokens.append({
-            'type': 'block_html',
-            'pre': pre,
-            'text': text
-        })
+        tag = m.group(1)
+        if not tag:
+            text = m.group(0)
+            self.tokens.append({
+                'type': 'close_html',
+                'text': text
+            })
+        else:
+            attr = m.group(2)
+            text = m.group(3)
+            self.tokens.append({
+                'type': 'open_html',
+                'tag': tag,
+                'extra': attr,
+                'text': text
+            })
 
     def parse_paragraph(self, m):
         text = m.group(1).rstrip('\n')
@@ -421,17 +467,20 @@ class InlineGrammar(object):
     """Grammars for inline level tokens."""
 
     escape = re.compile(r'^\\([\\`*{}\[\]()#+\-.!_>~|])')  # \* \+ \! ....
-    tag = re.compile(
-        r'^<!--[\s\S]*?-->|'  # comment
-        r'^<\/\w+>|'  # close tag
-        r'^<\w+[^>]*?>'  # open tag
+    inline_html = re.compile(
+        r'^(?:%s|%s|%s)' % (
+            r'<!--[\s\S]*?-->',
+            r'<(\w+%s)((?:%s)*?)\s*>([\s\S]*?)<\/\1>' % (
+                _valid_end, _valid_attr),
+            r'<\w+%s(?:%s)*?\s*\/?>' % (_valid_end, _valid_attr),
+        )
     )
-    autolink = re.compile(r'^<([^ >]+(@|:\/)[^ >]+)>')
+    autolink = re.compile(r'^<([^ >]+(@|:)[^ >]+)>')
     link = re.compile(
         r'^!?\[('
         r'(?:\[[^^\]]*\]|[^\[\]]|\](?=[^\[]*\]))*'
         r')\]\('
-        r'''\s*<?([\s\S]*?)>?(?:\s+['"]([\s\S]*?)['"])?\s*'''
+        r'''\s*(<)?([\s\S]*?)(?(2)>)(?:\s+['"]([\s\S]*?)['"])?\s*'''
         r'\)'
     )
     reflink = re.compile(
@@ -442,18 +491,18 @@ class InlineGrammar(object):
     nolink = re.compile(r'^!?\[((?:\[[^\]]*\]|[^\[\]])*)\]')
     url = re.compile(r'''^(https?:\/\/[^\s<]+[^<.,:;"')\]\s])''')
     double_emphasis = re.compile(
-        r'^_{2}(.+?)_{2}(?!_)'  # __word__
+        r'^_{2}([\s\S]+?)_{2}(?!_)'  # __word__
         r'|'
-        r'^\*{2}(.+?)\*{2}(?!\*)'  # **word**
+        r'^\*{2}([\s\S]+?)\*{2}(?!\*)'  # **word**
     )
     emphasis = re.compile(
-        r'^\b_((?:__|.)+?)_\b'  # _word_
+        r'^\b_((?:__|[^_])+?)_\b'  # _word_
         r'|'
-        r'^\*((?:\*\*|.)+?)\*(?!\*)'  # *word*
+        r'^\*((?:\*\*|[^\*])+?)\*(?!\*)'  # *word*
     )
-    code = re.compile(r'^(`+)\s*(.*?[^`])\s*\1(?!`)')  # `code`
+    code = re.compile(r'^(`+)\s*([\s\S]*?[^`])\s*\1(?!`)')  # `code`
     linebreak = re.compile(r'^ {2,}\n(?!\s*$)')
-    strikethrough = re.compile(r'^~~(?=\S)(.*?\S)~~')  # ~~word~~
+    strikethrough = re.compile(r'^~~(?=\S)([\s\S]*?\S)~~')  # ~~word~~
     footnote = re.compile(r'^\[\^([^\]]+)\]')
     text = re.compile(r'^[\s\S]+?(?=[\\<!\[_*`~]|https?://| {2,}\n|$)')
 
@@ -472,11 +521,16 @@ class InlineLexer(object):
     grammar_class = InlineGrammar
 
     default_rules = [
-        'escape', 'autolink', 'url', 'tag',
+        'escape', 'inline_html', 'autolink', 'url',
         'footnote', 'link', 'reflink', 'nolink',
         'double_emphasis', 'emphasis', 'code',
         'linebreak', 'strikethrough', 'text',
     ]
+    inline_html_rules = [
+        'escape', 'inline_html', 'autolink', 'url', 'link', 'reflink',
+        'nolink', 'double_emphasis', 'emphasis', 'code',
+        'linebreak', 'strikethrough', 'text',
+    ]
 
     def __init__(self, renderer, rules=None, **kwargs):
         self.renderer = renderer
@@ -487,13 +541,18 @@ class InlineLexer(object):
         if not rules:
             rules = self.grammar_class()
 
+        kwargs.update(self.renderer.options)
+        if kwargs.get('hard_wrap'):
+            rules.hard_wrap()
+
         self.rules = rules
 
         self._in_link = False
         self._in_footnote = False
+        self._parse_inline_html = kwargs.get('parse_inline_html')
 
-    def __call__(self, text):
-        return self.output(text)
+    def __call__(self, text, rules=None):
+        return self.output(text, rules)
 
     def setup(self, links, footnotes):
         self.footnote_index = 0
@@ -520,24 +579,23 @@ class InlineLexer(object):
                 out = getattr(self, 'output_%s' % key)(m)
                 if out is not None:
                     return m, out
-            return False
+            return False  # pragma: no cover
 
-        self.line_started = False
         while text:
             ret = manipulate(text)
-            self.line_started = True
             if ret is not False:
                 m, out = ret
                 output += out
                 text = text[len(m.group(0)):]
                 continue
-            if text:
+            if text:  # pragma: no cover
                 raise RuntimeError('Infinite loop at: %s' % text)
 
         return output
 
     def output_escape(self, m):
-        return m.group(1)
+        text = m.group(1)
+        return self.renderer.escape(text)
 
     def output_autolink(self, m):
         link = m.group(1)
@@ -553,14 +611,21 @@ class InlineLexer(object):
             return self.renderer.text(link)
         return self.renderer.autolink(link, False)
 
-    def output_tag(self, m):
-        text = m.group(0)
-        lower_text = text.lower()
-        if lower_text.startswith('<a '):
-            self._in_link = True
-        if lower_text.startswith('</a>'):
-            self._in_link = False
-        return self.renderer.tag(text)
+    def output_inline_html(self, m):
+        tag = m.group(1)
+        if self._parse_inline_html and tag in _inline_tags:
+            text = m.group(3)
+            if tag == 'a':
+                self._in_link = True
+                text = self.output(text, rules=self.inline_html_rules)
+                self._in_link = False
+            else:
+                text = self.output(text, rules=self.inline_html_rules)
+            extra = m.group(2) or ''
+            html = '<%s%s>%s</%s>' % (tag, extra, text, tag)
+        else:
+            html = m.group(0)
+        return self.renderer.inline_html(html)
 
     def output_footnote(self, m):
         key = _keyify(m.group(1))
@@ -573,7 +638,7 @@ class InlineLexer(object):
         return self.renderer.footnote_ref(key, self.footnote_index)
 
     def output_link(self, m):
-        return self._process_link(m, m.group(2), m.group(3))
+        return self._process_link(m, m.group(3), m.group(4))
 
     def output_reflink(self, m):
         key = _keyify(m.group(2) or m.group(1))
@@ -787,6 +852,15 @@ class Renderer(object):
     def text(self, text):
         """Rendering unformatted text.
 
+        :param text: text content.
+        """
+        if self.options.get('parse_block_html'):
+            return text
+        return escape(text)
+
+    def escape(self, text):
+        """Rendering escape sequence.
+
         :param text: text content.
         """
         return escape(text)
@@ -797,7 +871,7 @@ class Renderer(object):
         :param link: link content or email address.
         :param is_email: whether this is an email or not.
         """
-        text = link = escape(link)
+        text = link = escape_link(link)
         if is_email:
             link = 'mailto:%s' % link
         return '<a href="%s">%s</a>' % (link, text)
@@ -809,8 +883,7 @@ class Renderer(object):
         :param title: title content for `title` attribute.
         :param text: text content for description.
         """
-        if link.startswith('javascript:'):
-            link = ''
+        link = escape_link(link)
         if not title:
             return '<a href="%s">%s</a>' % (link, text)
         title = escape(title, quote=True)
@@ -823,8 +896,7 @@ class Renderer(object):
         :param title: title text of the image.
         :param text: alt text of the image.
         """
-        if src.startswith('javascript:'):
-            src = ''
+        src = escape_link(src)
         text = escape(text, quote=True)
         if title:
             title = escape(title, quote=True)
@@ -835,13 +907,11 @@ class Renderer(object):
             return '%s />' % html
         return '%s>' % html
 
-    def tag(self, html):
-        """Rendering span level html tag.
+    def inline_html(self, html):
+        """Rendering span level pure html content.
 
-        :param html: html tag snippet.
+        :param html: text content of the html snippet.
         """
-        if self.options.get('skip_html'):
-            return ''
         if self.options.get('escape'):
             return escape(html)
         return html
@@ -858,7 +928,7 @@ class Renderer(object):
         """
         html = (
             '<sup class="footnote-ref" id="fnref-%s">'
-            '<a href="#fn-%s" rel="footnote">%d</a></sup>'
+            '<a href="#fn-%s">%d</a></sup>'
         ) % (escape(key), escape(key), index)
         return html
 
@@ -869,7 +939,7 @@ class Renderer(object):
         :param text: text content of the footnote.
         """
         back = (
-            '<a href="#fnref-%s" rev="footnote">&#8617;</a>'
+            '<a href="#fnref-%s" class="footnote">&#8617;</a>'
         ) % escape(key)
         text = text.rstrip()
         if text.endswith('</p>'):
@@ -892,11 +962,14 @@ class Markdown(object):
     """The Markdown parser.
 
     :param renderer: An instance of ``Renderer``.
+    :param inline: An inline lexer class or instance.
+    :param block: A block lexer class or instance.
     """
-
     def __init__(self, renderer=None, inline=None, block=None, **kwargs):
         if not renderer:
             renderer = Renderer(**kwargs)
+        else:
+            kwargs.update(renderer.options)
 
         self.renderer = renderer
 
@@ -908,16 +981,15 @@ class Markdown(object):
         if inline:
             self.inline = inline
         else:
-            rules = InlineGrammar()
-            if kwargs.get('hard_wrap'):
-                rules.hard_wrap()
-            self.inline = InlineLexer(renderer, rules=rules)
+            self.inline = InlineLexer(renderer, **kwargs)
 
         self.block = block or BlockLexer(BlockGrammar())
-        self.options = kwargs
         self.footnotes = []
         self.tokens = []
 
+        # detect if it should parse text in block html
+        self._parse_block_html = kwargs.get('parse_block_html')
+
     def __call__(self, text):
         return self.parse(text)
 
@@ -968,7 +1040,7 @@ class Markdown(object):
     def peek(self):
         if self.tokens:
             return self.tokens[-1]
-        return None
+        return None  # pragma: no cover
 
     def output(self, text, rules=None):
         self.tokens = self.block(text, rules)
@@ -1079,12 +1151,19 @@ class Markdown(object):
         self.inline._in_footnote = False
         return self.renderer.placeholder()
 
-    def output_block_html(self):
+    def output_close_html(self):
         text = self.token['text']
-        if self.options.get('parse_html') and not self.token.get('pre'):
-            text = self.inline(text)
         return self.renderer.block_html(text)
 
+    def output_open_html(self):
+        text = self.token['text']
+        tag = self.token['tag']
+        if self._parse_block_html and tag not in _pre_tags:
+            text = self.inline(text, rules=self.inline.inline_html_rules)
+        extra = self.token.get('extra') or ''
+        html = '<%s%s>%s</%s>' % (tag, extra, text, tag)
+        return self.renderer.block_html(html)
+
     def output_paragraph(self):
         return self.renderer.paragraph(self.inline(self.token['text']))
 
@@ -1092,11 +1171,14 @@ class Markdown(object):
         return self.renderer.paragraph(self.tok_text())
 
 
-def markdown(text, **kwargs):
+def markdown(text, escape=True, **kwargs):
     """Render markdown formatted text to html.
 
     :param text: markdown formatted text content.
-    :param escape: if set to True, all html tags will be escaped.
+    :param escape: if set to False, all html tags will not be escaped.
     :param use_xhtml: output with xhtml tags.
+    :param hard_wrap: if set to True, it will use the GFM line breaks feature.
+    :param parse_block_html: parse text only in block level html.
+    :param parse_inline_html: parse text only in inline level html.
     """
-    return Markdown(**kwargs)(text)
+    return Markdown(escape=escape, **kwargs)(text)
